/*
 *  scsi_windows.cpp - SCSI Manager for Win32
 *
 *  Basilisk II (C) 1997-1999 Christian Bauer
 *
 *  Ported to Windows by Lauri Pesonen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// TODO: immediate bit handling.
// Now everything is synchronous
// what's scsi command 0x06? Iomega Zip uses that

#include "windows.h"
#include "wnaspi32.h"
#include "scsidefs.h"
#include "sysdeps.h"
#include "prefs.h"
#include "scsi.h"
#include "user_strings.h"
#include "main_windows.h"
#include "main.h"

#define DEBUG 0
#include "debug.h"


#define SCSI_READ_DISC_INFORMATION  0x51
#define SCSI_READ_TRACK_INFORMATION 0x52
#define SCSI_CLOSE_TRACK_OR_SESSION 0x5B
#define SCSI_BLANK_DISC							0xA1
#define SCSI_SET_SPEED							0xBB

#define NEC_CDROM_TOC_SIZE          1022
#define SIZE_INQUIRYBUF							36

typedef struct {
	int haid, target, lun;
	DWORD max_transfer;
} aspi_id;


// Max cmd size ... heh 12 would probably suffice
#define MAXCOMMAND 100
static int scsi_command_length = 0;
static BYTE scsi_command[MAXCOMMAND];

// This is the max buffer size
DWORD dwMaxTransfer;

// map Mac (target,lun) -> Windows (ha,target,lun)
// uses indices 1..6 as for now
static aspi_id fds[8*8];
static aspi_id scsi_target;

static BOOL is_aspi_available = false;

static int buffer_size = 0;			// Size of data buffer
static BYTE *buffer = NULL;			// Pointer to data buffer


#define MOVESCSIDWORD(pdwSrc,pdwDst) \
{\
    ((PBYTE)(pdwDst))[0] = ((PBYTE)(pdwSrc))[3];\
    ((PBYTE)(pdwDst))[1] = ((PBYTE)(pdwSrc))[2];\
    ((PBYTE)(pdwDst))[2] = ((PBYTE)(pdwSrc))[1];\
    ((PBYTE)(pdwDst))[3] = ((PBYTE)(pdwSrc))[0];\
}

#define MOVESCSIWORD(pwSrc,pwDst) \
{\
    ((PBYTE)(pwDst))[0] = ((PBYTE)(pwSrc))[1];\
    ((PBYTE)(pwDst))[1] = ((PBYTE)(pwSrc))[0];\
}

BOOL       gbASPIBuffer = FALSE;           // TRUE if ASPI transfer buffer
ASPI32BUFF gab;                            // Pointer to transfer buffer
HINSTANCE  ghinstWNASPI32 = 0;             // Handle to ASPI library

#define ASPICALL _cdecl

typedef DWORD (ASPICALL *LPGETASPI32SUPPORTINFO)(VOID);
typedef DWORD (ASPICALL *LPSENDASPI32COMMAND)(LPSRB);
typedef BOOL (ASPICALL *LPGETASPI32BUFFER)(PASPI32BUFF);
typedef BOOL (ASPICALL *LPFREEASPI32BUFFER)(PASPI32BUFF);
typedef BOOL (ASPICALL *LPTRANSLATEASPI32ADDRESS)(PDWORD,PDWORD);

DWORD (ASPICALL *gpfnGetASPI32SupportInfo)( VOID );
DWORD (ASPICALL *gpfnSendASPI32Command)( LPSRB );
BOOL  (ASPICALL *gpfnGetASPI32Buffer)( PASPI32BUFF );
BOOL  (ASPICALL *gpfnFreeASPI32Buffer)( PASPI32BUFF );
BOOL  (ASPICALL *gpfnTranslateASPI32Address)( PDWORD, PDWORD );

#define VENDOR_LEN 8
#define PRODUCT_LEN 16

struct smap_type {
	char vendor_from[VENDOR_LEN+1];
	char product_from[PRODUCT_LEN+1];
	char vendor_to[VENDOR_LEN+1];
	char product_to[PRODUCT_LEN+1];
	struct smap_type *next;
};

static struct smap_type *smaps = 0;
static struct smap_type *sdisable = 0;

static void cleanup_inquiry_name( char *name, int len )
{
  int  i;

  for( i = len; i--; ) {
    if( name[i] != 0 && name[i] != ' ' ) {
      break;
    }
    name[i] = 0;
  }
  name[len] = 0;
}

static void inquiry_2_name( LPBYTE lpInquiryBuf, LPSTR name )
{
  CHAR szVendor[VENDOR_LEN+1];
  CHAR szProduct[PRODUCT_LEN+1];

  memcpy( szVendor, lpInquiryBuf + 8, VENDOR_LEN );
	cleanup_inquiry_name( szVendor, VENDOR_LEN );

  memcpy( szProduct, lpInquiryBuf + 16, PRODUCT_LEN );
	cleanup_inquiry_name( szProduct, PRODUCT_LEN );

  wsprintf( name, "%s %s", szVendor, szProduct );
}

VOID aspi_final( VOID )
{
  if( gab.AB_BufPointer ) {
    if( gbASPIBuffer ) {
      gpfnFreeASPI32Buffer( &gab );
    } else {
      VirtualFree( gab.AB_BufPointer, 0, MEM_RELEASE );        
    }
		gab.AB_BufPointer = 0;
  }
	if(buffer) {
		VirtualFree( buffer, 0, MEM_RELEASE );        
		buffer = 0;
		buffer_size = 0;
	}
  if( ghinstWNASPI32 ) {
    FreeLibrary( ghinstWNASPI32 );
		ghinstWNASPI32 = 0;
  }
}

static BOOL aspi_init( VOID )
{
  UINT                fuPrevErrorMode;
  DWORD               dwSupportInfo;
  DWORD               dwLastError;
  SRB_GetSetTimeouts  srbTimeouts;
  OSVERSIONINFO				osvi;

	gab.AB_BufPointer = 0;
	ghinstWNASPI32 = 0;

  osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
  GetVersionEx( &osvi );
  if( osvi.dwPlatformId == VER_PLATFORM_WIN32s ) {
		ErrorAlert( "This program does not run under Win32s." );
    return( FALSE );
  }

  fuPrevErrorMode = SetErrorMode( SEM_NOOPENFILEERRORBOX );
  ghinstWNASPI32 = LoadLibrary( "WNASPI32" );
  dwLastError = GetLastError();
  SetErrorMode( fuPrevErrorMode );
  if( !ghinstWNASPI32 ) {
		char err[200];
		wsprintf( err, "Could not load ASPI layer, error code %d", dwLastError );
		ErrorAlert( err );
    return( FALSE );
  }

  gpfnGetASPI32SupportInfo = (LPGETASPI32SUPPORTINFO)GetProcAddress( ghinstWNASPI32, "GetASPI32SupportInfo" );
  gpfnSendASPI32Command = (LPSENDASPI32COMMAND)GetProcAddress( ghinstWNASPI32, "SendASPI32Command" );
  gpfnGetASPI32Buffer = (LPGETASPI32BUFFER)GetProcAddress( ghinstWNASPI32, "GetASPI32Buffer" );
  gpfnFreeASPI32Buffer = (LPFREEASPI32BUFFER)GetProcAddress( ghinstWNASPI32, "FreeASPI32Buffer" );
  gpfnTranslateASPI32Address = (LPTRANSLATEASPI32ADDRESS)GetProcAddress( ghinstWNASPI32, "TranslateASPI32Address" );

  if( !gpfnGetASPI32SupportInfo || !gpfnSendASPI32Command ) {
    FreeLibrary( ghinstWNASPI32 );
		ghinstWNASPI32 = 0;
		ErrorAlert( "Could not find required ASPI entry points." );
    return FALSE;
  }

  dwSupportInfo = gpfnGetASPI32SupportInfo();
  if( HIBYTE(LOWORD(dwSupportInfo)) != SS_COMP && 
			HIBYTE(LOWORD(dwSupportInfo)) != SS_NO_ADAPTERS )
  {
    FreeLibrary( ghinstWNASPI32 );
		ghinstWNASPI32 = 0;
		ErrorAlert( "Error in loading ASPI library." );
    return FALSE;
  }

  memset( &srbTimeouts, 0, sizeof(SRB_GetSetTimeouts) );
  srbTimeouts.SRB_Cmd = SC_GETSET_TIMEOUTS;
  srbTimeouts.SRB_HaId = 0xFF;
  srbTimeouts.SRB_Flags = SRB_DIR_OUT;
  srbTimeouts.SRB_Target = 0xFF;
  srbTimeouts.SRB_Lun = 0xFF;
  srbTimeouts.SRB_Timeout = 10*2; // timeout is in half seconds
  gpfnSendASPI32Command( &srbTimeouts );

  gbASPIBuffer = gpfnGetASPI32Buffer ? TRUE : FALSE;

  if( gbASPIBuffer ) {
    gbASPIBuffer = FALSE;
    for( dwMaxTransfer = 524288lu; dwMaxTransfer > 65536lu; dwMaxTransfer >>= 1 ) {
      memset( &gab, 0, sizeof(ASPI32BUFF) );
      gab.AB_BufLen = dwMaxTransfer;
      gab.AB_ZeroFill = FALSE;
      gbASPIBuffer = gpfnGetASPI32Buffer( &gab );
      if( gbASPIBuffer ) break;
    }
  }

  if( !gbASPIBuffer ) {
    dwMaxTransfer = gab.AB_BufLen = 65536lu;
    gab.AB_BufPointer = (PBYTE)VirtualAlloc (
        NULL,
        gab.AB_BufLen,
        MEM_RESERVE|MEM_COMMIT,
        PAGE_READWRITE
    );
  }

  if( !gab.AB_BufPointer ) {
    FreeLibrary( ghinstWNASPI32 );
		ghinstWNASPI32 = 0;
		ErrorAlert( "Out of memory allocating transfer buffer." );
    return FALSE;
  }

  return( TRUE );
}

// TODO:
//   inspect "immediate" bit of some important scsi commands
//   and if set, do not wait on event, but set up a completion callback

static bool sync_exec(
	BYTE            byHaId,
	BYTE            byTarget,
  BYTE            byFlags,
  DWORD           dwBufferBytes,
  PBYTE           pbyBuffer,
  BYTE            byCDBBytes,
  PBYTE           pbyCDB,
	uint16					*stat,
	BOOL						silent
)
{
  BOOL            bRetry = TRUE;
  DWORD           dwASPIStatus;
  HANDLE          heventExec;

	// May need larger sense area...? unbelievable
  // SRB_ExecSCSICmd srbExec;
	char sbuf[ sizeof(SRB_ExecSCSICmd) + 256 ];
  SRB_ExecSCSICmd *psrbExec = (SRB_ExecSCSICmd *)sbuf;

  memset( psrbExec, 0, sizeof(sbuf) );
  memcpy( psrbExec->CDBByte, pbyCDB, byCDBBytes );
  psrbExec->SRB_Cmd = SC_EXEC_SCSI_CMD;
  psrbExec->SRB_HaId = byHaId;
  psrbExec->SRB_Flags = byFlags;
  psrbExec->SRB_Target = byTarget;
  psrbExec->SRB_BufLen = dwBufferBytes;
  psrbExec->SRB_BufPointer = pbyBuffer;
	// Max reliable sense length under NT us 18.
  psrbExec->SRB_SenseLen = 18; // SENSE_LEN;
  psrbExec->SRB_CDBLen = byCDBBytes;

	if(stat) *stat = STATUS_GOOD;

  RetryExec:
  {
    heventExec = CreateEvent( NULL, TRUE, FALSE, NULL );
    if( heventExec ) {
      psrbExec->SRB_Flags |= SRB_EVENT_NOTIFY;
      psrbExec->SRB_PostProc = (LPVOID)heventExec;
      dwASPIStatus = gpfnSendASPI32Command( (LPSRB)psrbExec );
      if( dwASPIStatus == SS_PENDING ) {
        WaitForSingleObject( heventExec, INFINITE );
      }
			// if(psrbExec->SRB_Status == SS_PENDING) psrbExec->SRB_Status = SS_COMP;
      CloseHandle( heventExec );
    } else {
			// Cant create an event, wait in a busy loop
      gpfnSendASPI32Command( (LPSRB)psrbExec );
      while( psrbExec->SRB_Status == SS_PENDING );
    }
  }

	if(stat) *stat = psrbExec->SRB_TargStat;

  // Retry on the first unit attention.  Anything else will generate an error dialog.
	// TODO: is this check really ok.
  if( psrbExec->SRB_Status != SS_COMP ) {
    if( bRetry && 
        (psrbExec->SRB_TargStat != STATUS_CHKCOND ||
        (psrbExec->SenseArea[2] & 0x0F) != KEY_UNITATT) )
    {
      bRetry = FALSE;
      goto RetryExec;
    } else {
		  if(!silent) {
				char err[200];
				wsprintf( err, "SCSI io failure." );
				ErrorAlert( err );
			}
      return FALSE;
    }
  }
  return TRUE;    
}

VOID aspi_rescan_bus( VOID )
{
  BYTE                    byHaId;
  BYTE                    byMaxHaId;
  DWORD                   dwASPIStatus;
  SRB_RescanPort          srbRescanPort;

  dwASPIStatus = gpfnGetASPI32SupportInfo();
  if( HIBYTE(LOWORD(dwASPIStatus)) == SS_COMP ) {
    byMaxHaId = LOBYTE(LOWORD(dwASPIStatus));
    for( byHaId = 0; byHaId < byMaxHaId; byHaId++ ) {
      memset( &srbRescanPort, 0, sizeof(SRB_RescanPort) );
      srbRescanPort.SRB_Cmd = SC_RESCAN_SCSI_BUS;
      srbRescanPort.SRB_HaId = byHaId;
      gpfnSendASPI32Command( (LPSRB)&srbRescanPort );
    }
  }
}

static void clear_map(void)
{
  BYTE byTarget, byLun;

	scsi_command_length = 0;
	memset( scsi_command, 0, MAXCOMMAND );

	for(byTarget=0; byTarget<8; byTarget++) {
		for(byLun=0; byLun<8; byLun++) {
			fds[byTarget * 8 + byLun].haid = -1;
			fds[byTarget * 8 + byLun].target = -1;
			fds[byTarget * 8 + byLun].lun = -1;
		}
	}
	scsi_target.haid = -1;
	scsi_target.target = -1;
	scsi_target.lun = -1;
	scsi_target.max_transfer = -1;
}

static void add_mapping( struct smap_type *pmap, struct smap_type **root )
{
	struct smap_type *newmap = (struct smap_type *)malloc( sizeof(struct smap_type) );
	if(newmap) {
		memcpy( newmap, pmap, sizeof(struct smap_type) );
		newmap->next = *root;
		*root = newmap;
	}
}

static void delete_mappings( void )
{
	struct smap_type *tmp;

	while(smaps) {
		tmp = smaps;
		smaps = smaps->next;
		free(tmp);
	}
	smaps = 0;
	while(sdisable) {
		tmp = sdisable;
		sdisable = sdisable->next;
		free(tmp);
	}
	sdisable = 0;
}


static bool get_token( const char *line, int index, char *str, int maxlen )
{
	int len, quotes = index * 2 + 1;
	char *end;

	for( int i=0; i<quotes; i++ ) {
		line = strchr(line,'\"');
		if(!line) return(false);
		line++;
	}
	end = strchr(line,'\"');
	if(!end) return(false);
	len = (uint32)end - (uint32)line;
	if(len > maxlen) return(false);
	memcpy( str, line, len );
	str[len] = 0;
	return(true);
}

/* prefs:
noscsi true
disablescsi "TOSHIBA" "CD-ROM XM-6202B"
replacescsi "HP" "CD-Writer+ 7100" "PHILIPS" "CDD3600"
*/

static void read_scsi_mappings(void)
{
	int32 index;
	const char *str;
	struct smap_type map;

	index = 0;
	while ((str = PrefsFindString("replacescsi", index++)) != NULL) {
		memset( &map, 0, sizeof(map) );
		if( get_token(str,0,map.vendor_from,VENDOR_LEN) &&
				get_token(str,1,map.product_from,PRODUCT_LEN) &&
				get_token(str,2,map.vendor_to,VENDOR_LEN) &&
				get_token(str,3,map.product_to,PRODUCT_LEN) )
		{
			add_mapping( &map, &smaps );
		}
	}

	index = 0;
	while ((str = PrefsFindString("disablescsi", index++)) != NULL) {
		memset( &map, 0, sizeof(map) );
		if( get_token(str,0,map.vendor_from,VENDOR_LEN) &&
				get_token(str,1,map.product_from,PRODUCT_LEN) )
		{
			add_mapping( &map, &sdisable );
		}
	}
}

// Impersonate some devices
static void patch_inquiry( char *lpInquiryBuf )
{
	char vendor[VENDOR_LEN+1];
	char product[PRODUCT_LEN+1];
	struct smap_type *s = smaps;

	char *vp = lpInquiryBuf + 8;
	char *pp = lpInquiryBuf + 16;

	memcpy( vendor, vp, VENDOR_LEN );
	memcpy( product, pp, PRODUCT_LEN );

	cleanup_inquiry_name( vendor, VENDOR_LEN );
	cleanup_inquiry_name( product, PRODUCT_LEN );

	while(s) {
		if(strcmp(vendor,s->vendor_from) == 0) {
			if(strcmp(product,s->product_from) == 0) {
				memset(vp,' ',VENDOR_LEN);
				memset(pp,' ',PRODUCT_LEN);
				memcpy(vp,s->vendor_to,strlen(s->vendor_to));
				memcpy(pp,s->product_to,strlen(s->product_to));
				break;
			}
		}
		s = s->next;
	}
}

static bool is_disabled( char *lpInquiryBuf )
{
	char vendor[VENDOR_LEN+1];
	char product[PRODUCT_LEN+1];
	struct smap_type *s = sdisable;

	char *vp = lpInquiryBuf + 8;
	char *pp = lpInquiryBuf + 16;

	memcpy( vendor, vp, VENDOR_LEN );
	memcpy( product, pp, PRODUCT_LEN );

	cleanup_inquiry_name( vendor, VENDOR_LEN );
	cleanup_inquiry_name( product, PRODUCT_LEN );

	while(s) {
		if(strcmp(vendor,s->vendor_from) == 0) {
			if(strcmp(product,s->product_from) == 0) {
				return(true);
			}
		}
		s = s->next;
	}
	return(false);
}

/*
 *  Initialization
 */

void SCSIInit(void)
{
  DWORD						dwASPIStatus;
  BYTE            byHaId;
  BYTE            byMaxHaId;
  BYTE            byTarget;
  BYTE            byMaxTarget;
  SRB_HAInquiry   srbHAInquiry;
  SRB_GDEVBlock   srbGDEVBlock;
  BYTE            byInquiryCDB[6];
  BYTE            byInquiryBuf[SIZE_INQUIRYBUF+50];
	DWORD						dwMaxTransferBytes;

	// Allocate SCSI id's 1..6, leave 0 and 7 alone.
	int next_index = 1, last_index = 6;

	// SCSI disabled?
	if(PrefsFindBool("noscsi")) return;

	read_scsi_mappings();

	is_aspi_available = aspi_init();

	if(is_aspi_available) {

		buffer_size = dwMaxTransfer;
    buffer = (PBYTE)VirtualAlloc ( NULL, buffer_size, MEM_RESERVE|MEM_COMMIT, PAGE_READWRITE );
		if (!buffer) {
			ErrorAlert(GetString(STR_NO_MEM_ERR));
			QuitEmulator();
		}

		// aspi_rescan_bus();
		clear_map();

		dwASPIStatus = gpfnGetASPI32SupportInfo();
		if( HIBYTE(LOWORD(dwASPIStatus)) == SS_COMP ) {
			byMaxHaId = LOBYTE(LOWORD(dwASPIStatus));
			for( byHaId = 0; byHaId < byMaxHaId; byHaId++ ) {
				memset( &srbHAInquiry, 0, sizeof(SRB_HAInquiry) );
				srbHAInquiry.SRB_Cmd = SC_HA_INQUIRY;
				srbHAInquiry.SRB_HaId = byHaId;
				gpfnSendASPI32Command( (LPSRB)&srbHAInquiry );
				if( srbHAInquiry.SRB_Status != SS_COMP ) continue;
				byMaxTarget = srbHAInquiry.HA_Unique[3];
				if( byMaxTarget != 8 && byMaxTarget != 16 ) byMaxTarget = 8;
				for( byTarget = 0; byTarget < byMaxTarget; byTarget++ ) {
					memset( &srbGDEVBlock, 0, sizeof(SRB_GDEVBlock) );
					srbGDEVBlock.SRB_Cmd = SC_GET_DEV_TYPE;
					srbGDEVBlock.SRB_HaId = byHaId;
					srbGDEVBlock.SRB_Target = byTarget;
					gpfnSendASPI32Command( (LPSRB)&srbGDEVBlock );

	        dwMaxTransferBytes = *(PDWORD)&srbHAInquiry.HA_Unique[4];
					if( dwMaxTransferBytes > gab.AB_BufLen ) dwMaxTransferBytes = gab.AB_BufLen;
					if( dwMaxTransferBytes < 65536lu ) dwMaxTransferBytes = 65536lu;

					memset( byInquiryCDB, 0, 6 );
					byInquiryCDB[0] = (BYTE)SCSI_INQUIRY;
					byInquiryCDB[4] = (BYTE)SIZE_INQUIRYBUF;
					BOOL bExec = sync_exec( 
							byHaId, byTarget, SRB_DIR_IN, 
							SIZE_INQUIRYBUF, byInquiryBuf,
							6, byInquiryCDB,
							NULL,
							TRUE
					);
				  
					// Test: disable all but vendor id IOMEGA
					// if(bExec && memcmp(&byInquiryBuf[8],"IOMEGA",6) != 0) bExec = FALSE;

					// Test: disable vendor id IOMEGA
					// if(bExec && memcmp(&byInquiryBuf[8],"IOMEGA",6) == 0) bExec = FALSE;

					if( bExec ) {
						char name[100];
						inquiry_2_name( byInquiryBuf, name );

						D(bug("SCSI device found: %s\n",name));
						if(!*name) {
							D(bug(" -- disabled (no name, something is wrong).\n"));
						} else if(is_disabled( (char*)byInquiryBuf )) {
							D(bug(" -- disabled by user.\n"));
						} else {
							// take only first lun now.
							if(next_index <= last_index) {
								fds[next_index * 8 + 0].haid = byHaId;
								fds[next_index * 8 + 0].target = byTarget;
								fds[next_index * 8 + 0].lun = 0;
								fds[next_index * 8 + 0].max_transfer = dwMaxTransferBytes;
								next_index++;
							} else {
								D(bug(" -- no room.\n"));
							}
						}
					}
				}
			}
		}
	} // if is_aspi_available

	// Reset SCSI bus
	SCSIReset();
}


/*
 *  Deinitialization
 */

void SCSIExit(void)
{
	if(is_aspi_available) {
		aspi_final();
		clear_map();
		is_aspi_available = false;
		delete_mappings();
	}
}


/*
 *  Set SCSI command to be sent by scsi_send_cmd()
 */

void scsi_set_cmd(int cmd_length, uint8 *cmd)
{
	if(is_aspi_available) {
		scsi_command_length = cmd_length;
		memcpy( scsi_command, cmd, min(cmd_length,MAXCOMMAND));
		if(cmd_length > MAXCOMMAND) {
			char str[256];
			wsprintf(str, "Too long SCSI command (%d)", cmd_length);
			ErrorAlert(str);
		}
	}
}


/*
 *  Check for presence of SCSI target
 */

bool scsi_is_target_present(int id)
{
	if(is_aspi_available) {
		// check first lun
		return fds[id * 8].haid >= 0;
	} else {
		return false;
	}
}


/*
 *  Set SCSI target (returns false on error)
 */

bool scsi_set_target(int id, int lun)
{
	if(is_aspi_available) {
		aspi_id new_fd = fds[id * 8 + lun];
		if (new_fd.haid < 0)
			return false;
		// Do not use auto sense.
		// if (memcmp(&new_fd,&scsi_target,sizeof(aspi_id)) != 0) {
		//   Clear sense data when selecting new target
		// }
		scsi_target = new_fd;
		return true;
	}
  return false;
}

static void get_scsi_command_name( BYTE c, char *cmd_name )
{
	switch(c) {
		case SCSI_INQUIRY:
			strcpy( cmd_name, "SCSI_INQUIRY" ); break;
    case SCSI_READ6:
      strcpy( cmd_name, "SCSI_READ6" ); break;
    case SCSI_REQ_SENSE:
      strcpy( cmd_name, "SCSI_REQ_SENSE" ); break;
    case SCSI_TST_U_RDY:
      strcpy( cmd_name, "SCSI_TST_U_RDY" ); break;
    case SCSI_FORMAT:
      strcpy( cmd_name, "SCSI_FORMAT" ); break;
    case SCSI_READ10:
      strcpy( cmd_name, "SCSI_READ10" ); break;
    case SCSI_RD_CAPAC:
      strcpy( cmd_name, "SCSI_RD_CAPAC" ); break;
    case SCSI_RELEASE:
      strcpy( cmd_name, "SCSI_RELEASE" ); break;
    case SCSI_SEND_DIAG:
      strcpy( cmd_name, "SCSI_SEND_DIAG" ); break;
    case SCSI_WRITE6:
      strcpy( cmd_name, "SCSI_WRITE6" ); break;
    case SCSI_WRITE10:
      strcpy( cmd_name, "SCSI_WRITE10" ); break;

    case SCSI_START_STP:
      strcpy( cmd_name, "SCSI_START_STP" ); break;
    case SCSI_SYNC_CACHE:
      strcpy( cmd_name, "SCSI_SYNC_CACHE" ); break;
    case SCSI_VERIFY:
      strcpy( cmd_name, "SCSI_VERIFY" ); break;
    case SCSI_MODE_SEL6:
      strcpy( cmd_name, "SCSI_MODE_SEL6" ); break;
    case SCSI_MODE_SEN6:
      strcpy( cmd_name, "SCSI_MODE_SEN6" ); break;
    case SCSI_MED_REMOVL:
      strcpy( cmd_name, "SCSI_MED_REMOVL" ); break;
			
    case SCSI_ERASE:
      strcpy( cmd_name, "SCSI_ERASE" ); break;
    case SCSI_RD_BLK_LIM:
      strcpy( cmd_name, "SCSI_RD_BLK_LIM" ); break;
    case SCSI_RESERVE:
      strcpy( cmd_name, "SCSI_RESERVE" ); break;
    case SCSI_REWIND:
      strcpy( cmd_name, "SCSI_REWIND" ); break;
    case SCSI_SPACE:
      strcpy( cmd_name, "SCSI_SPACE" ); break;
    case SCSI_WRT_FILE:
      strcpy( cmd_name, "SCSI_WRT_FILE" ); break;

    case SCSI_SETWINDOW:
      strcpy( cmd_name, "SCSI_SETWINDOW" ); break;

    case SCSI_READ_TOC:
      strcpy( cmd_name, "SCSI_READ_TOC" ); break;
    case SCSI_READ_DISC_INFORMATION:
      strcpy( cmd_name, "SCSI_READ_DISC_INFORMATION" ); break;
    case SCSI_READ_TRACK_INFORMATION:
      strcpy( cmd_name, "SCSI_READ_TRACK_INFORMATION" ); break;
    case SCSI_CLOSE_TRACK_OR_SESSION:
      strcpy( cmd_name, "SCSI_CLOSE_TRACK_OR_SESSION" ); break;
    case SCSI_SET_SPEED:
      strcpy( cmd_name, "SCSI_SET_SPEED" ); break;
    case SCSI_BLANK_DISC:
      strcpy( cmd_name, "SCSI_BLANK_DISC" ); break;
		default:
			strcpy( cmd_name, "command" );
	}
}

static void dbout_scsi_req(size_t data_length, bool reading, int sg_size, uint32 *sg_len)
{
	char msg[512], m2[20], cmd_name[100];
	int j;

	get_scsi_command_name( scsi_command[0], cmd_name );
	wsprintf( 
		msg,
		"%s %Xh bytes in %d s/g areas, host=(%d,%d,%d), %s[%d] = ",
		reading ? "reading" : "writing",
		(int)data_length,
		(int)sg_size,
		(int)scsi_target.haid,
		(int)scsi_target.target,
		(int)scsi_target.lun,
		cmd_name,
		(int)scsi_command_length
	);
	for( j=0; j<scsi_command_length; j++ ) {
		wsprintf( m2, "%02X ", (int)scsi_command[j] );
		strcat( msg, m2 );
	}
	strcat( msg, "\n" );
	OutputDebugString(msg);

	size_t data_length_verify = 0;
	for(j=0; j<sg_size; j++) {
		data_length_verify += sg_len[j];
	}
	if(data_length_verify != data_length) {
		wsprintf( msg, "data length mismatch: %02X != %02X\n", (int)data_length, (int)data_length_verify );
		OutputDebugString(msg);
	}
}

/*
 *  Check if requested data size fits into buffer, allocate new buffer if needed
 */

static bool try_buffer(int size)
{
	if (size <= buffer_size)
		return true;

  BYTE *new_buffer = (PBYTE)VirtualAlloc ( NULL, size, MEM_RESERVE|MEM_COMMIT, PAGE_READWRITE );
	if (new_buffer == NULL)
		return false;
  if(buffer) VirtualFree( buffer, 0, MEM_RELEASE );        
	buffer = new_buffer;
	buffer_size = size;
	return true;
}

// Splitting ASSUMES that it's a 10 byte read/write cdb
// with LBA @ index 2, len @ index 7.
// It is caller's responsibility to check this.
// In theory READ6 etc could need splitting too but I have never seen that

// split transfer. does some extra moving, but the mac requests
// are sometimes larger than aspi can handle
static bool sync_exec_split(
	BYTE            byHaId,
	BYTE            byTarget,
  BYTE            byFlags,
  DWORD           dwBufferBytes,
  PBYTE           pbyBuffer,
  BYTE            byCDBBytes,
  PBYTE           pbyCDB,
	uint16					*stat,
	BOOL						silent,
	DWORD						max_transfer
)
{
	bool retval = false;
	bool multi = false;
	uint32 block_size, block, n_blocks, bytes, bytes_left = dwBufferBytes;
	uint32 device_block_size, lba;
	uint16 device_blocks;
	
	block_size = min(max_transfer,dwMaxTransfer);

	n_blocks = (dwBufferBytes + block_size - 1) / block_size;
	if(n_blocks == 0) n_blocks = 1;

	if(n_blocks > 1) {
		multi = true;
		MOVESCSIDWORD( &pbyCDB[2], &lba );
		MOVESCSIWORD( &pbyCDB[7], &device_blocks );
		device_block_size = dwBufferBytes / device_blocks;
		D(bug("i/o: lba=%d, blocks=%d, block size=%d\n\n",(int)lba, (int)device_blocks, (int)device_block_size));
	}

	for( block=0; block<n_blocks; block++ ) {
		bytes = min(block_size, bytes_left);
		if(byFlags & SRB_DIR_OUT) {
			if(bytes) memcpy( gab.AB_BufPointer, pbyBuffer, bytes );
		}
		if(multi) {
			device_blocks = (uint16)(bytes / device_block_size);
			D(bug("i/o block %d of %d, lba=%d, blocks=%d\n\n",(int)block+1, (int)n_blocks, (int)lba, (int)device_blocks));
			MOVESCSIDWORD( &lba, &pbyCDB[2] );
			MOVESCSIWORD( &device_blocks, &pbyCDB[7] );
			lba += device_blocks;
		}
		retval = sync_exec(
			byHaId,
			byTarget,
			byFlags,
			bytes,
			gab.AB_BufPointer,
			byCDBBytes,
			pbyCDB,
			stat,
			silent
		);
		if(!retval) break;
		if(byFlags & SRB_DIR_IN) {
			if(bytes) memcpy( pbyBuffer, gab.AB_BufPointer, bytes );
		}
		pbyBuffer += bytes;
		bytes_left -= bytes;
	}
	return( retval );
}

/*
 *  Send SCSI command to active target (scsi_set_command() must have been called),
 *  read/write data according to S/G table (returns false on error)
 */

// The code assumes that data_length and scatter/gather total length match

bool scsi_send_cmd(size_t data_length, bool reading, int sg_size, uint8 **sg_ptr, uint32 *sg_len, uint16 *stat)
{
	bool retval = false;

	D(bug("SCSI_SEND_CMD START\n"));

	if(is_aspi_available && scsi_target.haid >= 0 && scsi_command_length > 0) {

		// Generally cannot patch just any cdb, the data length location varies
		if(data_length & 1) {
			dbout_scsi_req(data_length, reading, sg_size, sg_len);
		} else {
#if DEBUG
			dbout_scsi_req(data_length, reading, sg_size, sg_len);
#endif
		}

		// patches
		// note that this does not affect the emulation, only the
		// aspi layer and lower drivers
		switch(scsi_command[0]) {
			case SCSI_INQUIRY:
				if(data_length && data_length < SIZE_INQUIRYBUF) {
					scsi_command[4] = data_length = SIZE_INQUIRYBUF;
				} else if(data_length & 1) {
					scsi_command[4]++;
					data_length++;
				}
				break;
			case 0x06:
				// some downstream guy doesn't like byte aligned sizes.
				if(scsi_command[4] & 1) {
					scsi_command[4]++;
					data_length++;
				}
				break;
			case SCSI_READ_TOC:
				// some cd's don't work with partial toc requests
				if(data_length < NEC_CDROM_TOC_SIZE) {
					data_length = NEC_CDROM_TOC_SIZE;
					int16 src = data_length; // size_t may be 16 or 32 bits
					MOVESCSIWORD(&src,&scsi_command[7]);
				}
				break;
			default:
				break;
		}

		// Check if buffer is large enough, allocate new buffer if needed
		if (!try_buffer(data_length)) {
			char str[256];
			wsprintf(str, GetString(STR_SCSI_BUFFER_ERR), data_length);
			ErrorAlert(str);
			return false;
		}

		// Process S/G table when writing
		if (!reading) {
			D(bug(" writing to buffer\n"));
			uint8 *buffer_ptr = buffer;
			for (int i=0; i<sg_size; i++) {
				uint32 len = sg_len[i];
				D(bug("  %d bytes from %08lx\n", len, sg_ptr[i]));
				memcpy(buffer_ptr, sg_ptr[i], len);
				buffer_ptr += len;
			}
		}

		// Do not use auto sense.

		if(scsi_command[0] == SCSI_WRITE10 || scsi_command[0] == SCSI_READ10) {
			// These may need splitting.
			retval = sync_exec_split(
				scsi_target.haid,
				scsi_target.target,
				reading ? SRB_DIR_IN : SRB_DIR_OUT,
				data_length,
				buffer,
				scsi_command_length,
				scsi_command,
				stat,
				TRUE,
				scsi_target.max_transfer
			);
		} else {
			// Send regular command
			retval = sync_exec(
				scsi_target.haid,
				scsi_target.target,
				reading ? SRB_DIR_IN : SRB_DIR_OUT,
				data_length,
				buffer,
				scsi_command_length,
				scsi_command,
				stat,
				TRUE
			);
		}

		// Process S/G table when reading
		if (retval && reading) {

			if(scsi_command[0] == SCSI_INQUIRY) {
				patch_inquiry( (char*)buffer );
			}

			D(bug(" reading from buffer\n"));
			uint8 *buffer_ptr = buffer;
			for (int i=0; i<sg_size; i++) {
				uint32 len = sg_len[i];
				D(bug("  %d bytes to %08lx\n", len, sg_ptr[i]));
				memcpy(sg_ptr[i], buffer_ptr, len);
				buffer_ptr += len;
			}
		}

	} else {
		*stat = STATUS_GOOD;
		retval = false;
	}

	D(bug("SCSI_SEND_CMD END, retval = %d\n\n",(int)retval));

  return(retval);
}
